#!/usr/bin/python

program_name = "pyrpn"
program_version = "0.4"

cfg = None

#
# PyRPN
#
# Copyright (C) 2002 Scott Scriven
#
# Released under the terms of the GNU Public License
#

# gtk stuffs
# Takes a while to import gtk on some systems
print "Loading gtk..."
#from gtk import *
import gtk
print "Loading gtk...  done."

from tktk import ButtonChoiceWindow

def runqueue():
  while gtk.events_pending():
    gtk.mainiteration()

def ltoa(num, base):
  symbols = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  symbols = symbols[:base]

  text = ""
  div = long(base)
  while num > 0:
    mod = num % div
    text = symbols[mod] + text
    num = (num - mod) / div
  #text = "b%s: %s" % (base, repr(num))
  #text = text[:-1]
  # TODO: print in requested base
  #text = repr(num)
  #text = "base%s: %s" % (base, text)
  return text

class value:
  def __init__(self, data):
    # Actual value held by this object
    self.data = data
    # Displayable representation of object
    self.cached = ""
    # Datatype
    self.type = "generic"

  def str(self):
    if not self.cached:
      self.cached = repr(self.data)
    return self.cached

  def set(self, data):
    self.data = data
    self.cached = ""

class Int(value):
  def __init__(self, data):
    value.__init__(self, long(data))
    self.type = "int"
    # Number base to display in
    self.base = 10
    self.base_cached = 0

  def str(self):
    if not self.cached  or  self.base != self.base_cached:
      global cfg
      
      if self.base == 10:
        text = repr(self.data)
        text = text[:-1]
      else:
        text = ltoa(self.data, self.base)
      # add commas
      commas = cfg.commas
      comma = cfg.comma_type
      if commas:
        i = len(text) - commas
        while i > 0:
          text = text[:i] + comma + text[i:]
          i = i - commas
      if self.base != 10:
        text = "base" + repr(self.base) + ": " + text
      self.cached = text
      self.base_cached = self.base
    return self.cached

class Float(value):
  def __init__(self, data):
    value.__init__(self, float(data))
    self.type = "float"
    # TODO? set precision on a per-number basis?

  def str(self):
    if not self.cached:
      global cfg
      if cfg.precision > 0:
        format = "%%.%if" % (cfg.precision)
      else:
        format = "%s"
      text = format % (self.data)

      # add commas
      commas = cfg.commas
      comma = cfg.comma_type
      if commas:
        import string
        i = string.rfind(text, ".") - commas
        while i > 0:
          text = text[:i] + comma + text[i:]
          i = i - commas
      self.cached = text
    return self.cached

class Str(value):
  def __init__(self, data):
    value.__init__(self, data)
    self.type = "str"

class stack:
  """
  TODO: make sure all functions are "undo-safe" (?)
  
  Key:
    - Not done
    + Half done
    * Done
  
  Basic:
    * +, -, *, /
    * modulus
    - divmod
    * sin, cos, tan
    * asin, acos, atan
    * sinh, cosh, tanh
    + asinh, acosh, atanh (not implemented in python math module)
    * sqrt, y^x, 1/x; x^2, xroot, 10^x, log, e^x, ln
    * min / max
    * abs, sign
    - ipart, fpart
    * floor, ceil
    * rand, frand
    - EEX button
  
  Integer / Logic:
    * ->base (num, from, to) (up to base 36 (0-9,A-Z))
    * lshift, rshift
    - rroll, lroll (doesn't make sense with unlimited-size numbers!)
    - atonl, ntoal
    * and, or, xor, not
    * &, |, ^, ~
    * ->int, ->float
  
  Stack:
    * last / undo
    * drop
    * clear
    * dup
    - dupn, dropn
    * swap
    * roll
    + rolln (roll by x or -x items)
    * pick
    * edit
  
  Vector:
    - arrays, lists, matrices
    - vector operations...
  
  Complex:
    - complex numbers...
    - i
  
  Symbolic:
    - fractions
    - equations
    - value plugger
  
  Datatypes, units:
    * explicit datatyping
      * type
      * value
      - units
    + datatype conversion
      * ->int
      * ->float
      - ->string
      - ...
    * cached displayable values ?
    - unit conversion
      - deg, rad
      - inch, foot, yard, mile, mm, cm, m, km
      - g, kg, lbs, ...
      - custom units
      - ...
  
  Algebra:
    - eval expressions
  
  Statistics / probability:
    * combination
    * permutation
    * !
  
  Financial
  
  Constants:
    - e, i, pi, symbolic and numeric versions of each?
  
  Storage:
    - dir tab in middle area
    - show CWD somewhere...  window title?
    - add soft menus, for CST and files
    - directories, filesystem, etc...
      - implement as ~/.pyrpn/home ?
      - layer on top of /etc/pyrpn/home ?
        - create default files
    - sto
    - purge
    - home
    - up
    - edit file
    - distinguish better between strings and functions
    - find a way to determine what type each file is
      - foo.rpl (executable?): RPL script
      - foo.py (executable?): python script
      - foo.unit: unit
      - foo.pickle: pickled stack object?
      - foo.int: integer
      - foo.float: float
      - foo.txt: string
      - foo.whatever: whatever...
      - .foo: hidden, internal pyrpn use only
      - foo/: directory
      - CST/: custom menu (can be inherited)
  
  Programming:
    - RPL stack-based scripting...
      - {, }
      - if
      - for
      - while
    - python scripts
  
  Options / modes:
    * precision
    + degrees, radians, grads?
    + float modes
      * std (precision 0) 
      * fixed (precision n) 
      - sci floats
  """
  def __init__(self, cfg, parent=None, callback=None):
    self.data = []
    self.parent = parent
    self.cfg = cfg
    self.callback = callback
    if not callback:
      self.callback = self.nop

    self.functions = {
      "dump" : self.dump,

      "quit" : (self.callback, ["quit"]),
      "functions" : (self.callback, ["functions"]),
      
      "undo" : self.undo,
      "drop" : self.drop,
      "swap" : self.swap,
      "dup" : self.dup,
      "roll" : self.roll,
      "rolln" : self.rolln,
      "pick" : self.pick,
      "clear" : self.clear,
      "edit" : self.edit,
      
      "+" : self.add,
      "-" : self.subtract,
      "*" : self.multiply,
      "/" : self.divide,
      "neg" : self.negate,
      "%" : self.modulus,

      "min" : self.min,
      "max" : self.max,
      "abs" : (self.maths, ["abs"]),
      "sign" : self.sign,
      "floor" : (self.maths, ["floor"]),
      "ceil" : (self.maths, ["ceil"]),

      "rand" : self.rand,
      "frand" : self.frand,

      "perm" : self.perm,
      "comb" : self.comb,
      "!" : self.factorial,

      "->int" : self.to_int,
      "->float" : self.to_float,

      "shl" : self.shl,
      "shr" : self.shr,
      #"rol" : self.rol,
      #"ror" : self.ror,
      "&" : self.band,
      "|" : self.bor,
      "^" : self.bxor,
      "~" : self.bnot,
      "and" : self.iand,
      "or" : self.ior,
      "xor" : self.ixor,
      "not" : self.inot,
      
      "sqrt" : (self.maths, ["sqrt"]),
      "square" : self.square,
      "pow" : self.pow,
      "inv" : self.inv,
      "xroot" : self.xroot,
      "log" : (self.maths, ["log"]),
      "10pow" : self.pow10,
      "ln" : (self.maths, ["ln"]),
      "epow" : self.epow,
      
      "sin" : (self.maths, ["sin"]),
      "cos" : (self.maths, ["cos"]),
      "tan" : (self.maths, ["tan"]),
      "asin" : (self.maths, ["asin"]),
      "acos" : (self.maths, ["acos"]),
      "atan" : (self.maths, ["atan"]),
      "sinh" : (self.maths, ["sinh"]),
      "cosh" : (self.maths, ["cosh"]),
      "tanh" : (self.maths, ["tanh"]),
      "asinh" : (self.maths, ["asinh"]),
      "acosh" : (self.maths, ["acosh"]),
      "atanh" : (self.maths, ["atanh"]),
      
      "angle_unit" : self.angle_unit,
      "precision" : self.precision,
      "commas" : self.commas,
      "commatype" : self.comma_type,
      "base" : self.base,
      "->base" : self.to_base,
      }

  ##################################################
  #
  def execute(self, func):
    try:
      if func != self.undo:
        self.backup()
      if type(func) == type((1,2)):
        f = func[0]
        args = func[1]
        apply(f, args)
      else:
        func()
    except Exception:
      import sys
      ei = sys.exc_info()
      text = "%s: %s" % (ei[0], ei[1])
      text = text[11:]
      self.error("%s" % text)

  ##################################################
  #
  def operate(self, func, force=0):

    if self.functions.has_key(func)  or  force:
      self.execute(self.functions[func])

    else:
      self.push(func)

  ##################################################
  #
  def value(self, data):
    if type(data) in [type(1L), type(1)]:
      return Int(data)
    elif type(data) == type(1.0):
      return Float(data)
    elif type(data) == type("abc"):
      return Str(data)
    else:
      return value(data)

  ##################################################
  #
  def invalidate_cache(self):
    for d in self.data:
      d.cached = ""

  ##################################################
  #
  def dump(self):
    i = len(self.data) - 1
    while i >= 0:
      print "%s: %s (%s)" % (i+1, self.data[i].str(), self.data[i].data)
      i = i - 1

  ##################################################
  #
  def nop(self, *args):
    pass

  ##################################################
  # Return stack to previous state
  # (only works once)
  def backup(self):
    self.old_data = self.data[:]

  ##################################################
  # Return stack to previous state
  # (only works once)
  def undo(self):
    foo = self.old_data
    self.old_data = self.data
    self.data = foo

  ##################################################
  #
  def push(self, val):
    self.backup()
    data = self.value(val)
    self.data.insert(0, data)

  ##################################################
  #
  def error(self, text):
    title = "Stack Error"
    buttons = [("D'oh!", "ok")]
    ButtonChoiceWindow()(title, text, buttons, transient=self.parent)

  ##################################################
  #
  def drop(self):
    del self.data[0]

  ##################################################
  #
  def swap(self):
    a = self.data[0]
    b = self.data[1]
    self.data[0] = b
    self.data[1] = a
      
  ##################################################
  #
  def dup(self):
    a = self.data[0]
    self.push(a.data)
      
  ##################################################
  #
  def roll(self):
    self.rolln(1)
      
  ##################################################
  #
  def rolln(self, a=0):
    if not a:
      a = self.data[0]
      del self.data[0]

    l = len(self.data)
    if a >= l  or  a <= -l:
      raise ValueError, "Requested roll too large"

    if a > 0:
      slice = self.data[:a]
      del self.data[:a]
      self.data.extend(slice)
    else:
      slice = self.data[a:]
      del self.data[a:]
      self.data[:0] = slice
      
  ##################################################
  #
  def clear(self):
    self.data = []

  ##################################################
  #
  def edit(self):
    a = self.data[0]

    if a.type == "int":
      text = repr(a.data)
      text = text[:-1]
    elif a.type == "float":
      text = repr(a.data)
    elif a.type == "str":
      text = a.data
    else:
      text = repr(a.data)

    title = "%s (edit)" % (program_name)
    #import pyrpn_editor
    import pedit
    #edit = pyrpn_editor.Editor(cfg, None, text, title, "small")
    self.cfg.pedit_width = self.cfg.editor["small width"]
    self.cfg.pedit_height = self.cfg.editor["small height"]
    self.cfg.pedit_fonts = self.cfg.editor["fonts"]
    self.cfg.pedit_make_backups = self.cfg.editor["make backups"]
    self.cfg.pedit_fonts = self.cfg.editor["fonts"]
    edit = pedit.Editor(self.cfg, text=text, title=title, mode="mini", transient=self.parent)
    edit.run()
    text = edit.get_text()

    if a.type == "int":
      v = Int(text)
    elif a.type == "float":
      v = Float(text)
    elif a.type == "str":
      v = Str(text)
    else:
      v = value(text)

    self.data[0] = v
      
  ##################################################
  #
  def pick(self):
    a = self.data[0].data
    b = self.data[a]
    del self.data[a]
    self.data[0] = b


  ##################################################
  #
  def add(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data + a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def subtract(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data - a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def negate(self):
    a = self.data[0]
    a.set(-a.data)
    self.data[0] = a

  ##################################################
  #
  def multiply(self):
    a = self.data[0]
    b = self.data[1]

    # coerce floats to ints for string multiplication
    if a.type == "str":
      if b.type == "float":
        b.data = int(b.data)
    elif b.type == "str":
      if a.type == "float":
        a.data = int(a.data)
        
    c = self.value(b.data * a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def divide(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data / a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def modulus(self):
    a = self.data[0]
    b = self.data[1]
    if b.type == "float"  or  a.type == "float":
      import math
      c = Float(math.fmod(b.data, a.data))
      del self.data[1]
      self.data[0] = c
    elif b.type == "int"  and  a.type =="int":
      c = Int(b.data % a.data)
      del self.data[1]
      self.data[0] = c
    else:
      raise ValueError, "Invalid types for modulus"

  ##################################################
  #
  def min(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(min(a.data,b.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def max(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(max(a.data,b.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def sign(self):
    a = self.data[0]
    if a.data < 0:
      self.data[0] = Int(-1)
    else:
      self.data[0] = Int(1)

  ##################################################
  #
  def rand(self):
    import whrandom
    a = self.data[0]
    b = self.data[1]
    c = Int(whrandom.randint(b.data,a.data))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def frand(self):
    import whrandom
    self.push(whrandom.random())

  ##################################################
  #
  def perm(self):
    r = self.data[0]
    n = self.data[1]
    if n.data < r.data:
      raise ValueError, "Bad argument value"
    v = self._factorial(n.data) / self._factorial(n.data - r.data)
    del self.data[1]
    self.data[0] = Int(v)

  ##################################################
  #
  def comb(self):
    r = self.data[0]
    n = self.data[1]
    if n.data < r.data:
      raise ValueError, "Bad argument value"
    v = self._factorial(n.data) / \
        self._factorial(n.data - r.data) / \
        self._factorial(r.data)
    del self.data[1]
    self.data[0] = Int(v)

  ##################################################
  #
  def factorial(self):
    a = self.data[0]
    b = self._factorial(a.data)
    self.data[0] = Int(b)

  ##################################################
  #
  def _factorial(self, num):
    if not num > 0:
      raise ValueError, "Factorial takes a positive integer"
    b = 1L
    for c in xrange(num):
      b = b * (c + 1)
    return b

  ##################################################
  #
  def to_int(self):
    a = self.data[0]
    b = Int(a.data)
    self.data[0] = b

  ##################################################
  #
  def to_float(self):
    a = self.data[0]
    b = Float(a.data)
    self.data[0] = b

  ##################################################
  #
  def shl(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(b.data << a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def shr(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(b.data >> a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def band(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data & b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bor(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data | b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bxor(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(a.data ^ b.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def bnot(self):
    a = self.data[0]
    c = Int(~a.data)
    self.data[0] = c

  ##################################################
  #
  def iand(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(not(not(a.data and b.data)))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def ior(self):
    a = self.data[0]
    b = self.data[1]
    c = Int(not(not(a.data or b.data)))
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def ixor(self):
    a = self.data[0]
    b = self.data[1]
    c = 0
    if (a.data and (not b.data)) or \
       (b.data and (not a.data)):
      c = 1
    c = Int(c)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def inot(self):
    a = self.data[0]
    c = Int(not a.data)
    self.data[0] = c

  ##################################################
  # Execute simple one-argument functions from the
  # math module...
  # a(sin|cos|tan)h seem to be missing
  #
  def maths(self, which):
    import math
    trig = ["sin", "cos", "tan",
            "sinh", "cosh", "tanh"]
    atrig = ["asin", "acos", "atan",
            "asinh", "acosh", "atanh"]
    missing = ["asinh", "acosh", "atanh"]
    if which in missing:
      raise ValueError, "function not implemented"
    funcs = {
      "sin" : math.sin,
      "cos" : math.cos,
      "tan" : math.tan,
      "asin" : math.asin,
      "acos" : math.acos,
      "atan" : math.atan,
      "sinh" : math.sinh,
      "cosh" : math.cosh,
      "tanh" : math.tanh,
      #"asinh" : math.asinh,
      #"acosh" : math.acosh,
      #"atanh" : math.atanh,
      "sqrt" : math.sqrt,
      "log" : math.log10,
      "ln" : math.log,
      "abs" : abs,
      "floor" : math.floor,
      "ceil" : math.ceil,
      }
    a = self.data[0]
    if (self.cfg.angle_unit == "degrees")  and  (which in trig):
      a.set(a.data * math.pi / 180.0)
    func = funcs[which]
    a.set(func(a.data))
    if (self.cfg.angle_unit == "degrees")  and  (which in atrig):
      a.set(a.data * 180.0 / math.pi)
    self.data[0] = self.value(a.data)
      

  ##################################################
  #
  def square(self):
    import math
    a = self.data[0]
    self.data[0] = self.value(a.data*a.data)

  ##################################################
  #
  def pow(self):
    a = self.data[0]
    b = self.data[1]
    c = self.value(b.data ** a.data)
    del self.data[1]
    self.data[0] = c

  ##################################################
  #
  def inv(self):
    a = self.data[0]
    if a.type != "float":  # integers should become floats here
      self.to_float()
      a = self.data[0]
    b = self.value(1 / a.data)
    self.data[0] = b

  ##################################################
  #
  def xroot(self):
    a = Float(self.data[0].data)
    b = self.data[1]
    a.set(b.data ** (1/a.data))
    del self.data[1]
    self.data[0] = a

  ##################################################
  #
  def epow(self):
    import math
    a = self.data[0]
    self.data[0] = Float(math.e**a.data)

  ##################################################
  #
  def pow10(self):
    a = self.data[0]
    self.data[0] = Float(10.0**a.data)

  ##################################################
  #
  def angle_unit(self):
    a = self.data[0]
    if a.data in ["degrees", "deg"]:
      self.cfg.angle_unit = "degrees"
    elif a.data in ["radians", "rad"]:
      self.cfg.angle_unit = "radians"
    else:
      raise ValueError, "Invalid angle unit"

    del self.data[0]

  ##################################################
  #
  def precision(self):
    a = self.data[0]
    a = int(a.data)
    if a < 0:
      raise ValueError, "Precision must be non-negative integer"
    self.cfg.precision = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def commas(self):
    a = self.data[0]
    a = int(a.data)
    if a < 0:
      raise ValueError, "Commas must be non-negative integer"
    self.cfg.commas = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def comma_type(self):
    a = self.data[0]
    a = a.data
    self.cfg.comma_type = a

    del self.data[0]
    self.invalidate_cache()

  ##################################################
  #
  def base(self):
    a = self.data[0]
    a = int(a.data)
    if a < 2  or  a > 36:
      raise ValueError, "Base must be integer 2 ... 36"
    self.cfg.base = a

    del self.data[0]

  ##################################################
  #
  def to_base(self):
    a = self.data[0]
    b = self.data[1]
    a = int(a.data)
    if b.type != "int":
      raise ValueError, "Only integers can be converted to other bases"
    if a < 2  or  a > 36:
      raise ValueError, "Base must be integer 2 ... 36"
    b.base = a

    del self.data[0]

class calculator:
  """
  misc todo:
  - insert commas in numbers > 3 digits
  
  Possible frame/tab configuration:
  - basic: numpad, main ops
  - stack
  - alpha (fitaly)
  - trig
  - programming
  - matrix
  - units
  - file manager (always visible?)
  """

  def __init__(self, cfg):
    self.cfg = cfg

    self.win = gtk.Window()
    self.win.connect("destroy", self.destroy)
    self.win.set_title("%s %s" % (program_name, program_version))
    self.win.set_default_size(self.cfg.width,
                              self.cfg.height)

    self.stack = stack(self.cfg, self.win, self._stack_cb)

    self.vbox = gtk.VBox()
    self.win.add(self.vbox)

    self.make_menus()
    #self.make_vpaned()
    self.make_stacklist()
    self.make_tophalf()
    self.make_bottomhalf()

  def make_menus(self):
    self.ag = gtk.AccelGroup()
    self.win.add_accel_group(self.ag)
    self.itemf = gtk.ItemFactory(gtk.MenuBar, "<main>", self.ag)
    foo = [("/_Kill", None, None, 0, "<Branch>")]
    self.itemf.create_items([
      ("/File", None,  None,  0,  "<Branch>"),
      ("/File/_Quit", "<Control>Q",  self.destroy,  0,  ""),
      ("/Help", None,  None,  0,  "<Branch>"),
      ("/Help/About", None,  self.about_screen,  0,  ""),
      ("/Help/Functions", None,  self.functions_screen,  0,  ""),
      ])
    self.menubar = self.itemf.get_widget("<main>")
    self.vbox.pack_start(self.menubar, 0, 1)

    foo = gtk.ItemFactory(gtk.MenuBar, "<keydefs>", self.ag)
    foo.create_items([
      ("/foo",       None, None, 0, "<Branch>"),
      #("/foo/quit",  "<Control>Q", self.destroy, 0, ""),
      ])

  def make_stacklist(self):
    sw = gtk.ScrolledWindow()
    sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
    sw.set_placement(gtk.CORNER_BOTTOM_RIGHT)

    self.stack_lb = gtk.CList(2)
    #self.stack_lb.connect("select-row", self._stacklist_select_cb)
    self.stack_lb.set_selection_mode(gtk.SELECTION_BROWSE)
    self.stack_lb.set_column_justification(0, gtk.JUSTIFY_RIGHT)
    self.stack_lb.set_column_justification(1, gtk.JUSTIFY_RIGHT)
    self.stack_lb.set_column_auto_resize(0, 1)
    self.stack_lb.set_column_auto_resize(1, 1)
    sw.add(self.stack_lb)
    
    self.vbox.pack_start(sw, 1, 1)

    self.entry = gtk.Entry()
    self.entry.connect("activate", self._run_entry)
    self.vbox.pack_start(self.entry, 0, 0)

  def update_stack(self):
    self.stack_lb.freeze()
    self.stack_lb.clear()
    for i in xrange(len(self.stack.data)):
      val = self.stack.data[i]
      text = val.str()
      self.stack_lb.insert(0, ["%s:" % (i), text])

    self.stack_lb.thaw()

    # scrollbar needs to automatically scroll to the bottom
    a = self.stack_lb.get_vadjustment()
    if a.upper > a.page_size:
      scrollto = a.upper - a.page_size
      a.set_value(scrollto)
    self.entry.grab_focus()

  def _stack_cb(self, *args):
    funcs = {
      "quit" : self.destroy,
      "functions" : self.functions_screen,
      }
    if len(args) == 1:
      op = args[0]
      if funcs.has_key(op):
        funcs[op]()
    else:
      pass

  def make_vpaned(self):
    self.vpaned = gtk.VPaned()
    self.vbox.add(self.vpaned)

    self.vpane1 = gtk.VBox()
    self.vpane2 = gtk.VBox()
    self.vpaned.add(self.vpane1)
    self.vpaned.add(self.vpane2)

  def make_tophalf(self):
    nb = gtk.Notebook()
    self.vbox.pack_start(nb, 0, 1)

    t = self.make_exp()
    nb.append_page(t, gtk.Label("exp"))

    t = self.make_trig()
    nb.append_page(t, gtk.Label("trig"))

    #t = self.make_val()
    #nb.append_page(t, gtk.Label("val"))

    t = self.make_stat()
    nb.append_page(t, gtk.Label("stat"))

    t = self.make_int()
    nb.append_page(t, gtk.Label("int"))

    t = self.make_modes()
    nb.append_page(t, gtk.Label("modes"))

  def make_trig(self):
    t = gtk.Table(2,6)
    x = 0
    
    b = gtk.Button("sin")
    b.connect("clicked", self._operate, "sin")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("asin")
    b.connect("clicked", self._operate, "asin")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = gtk.Button("cos")
    b.connect("clicked", self._operate, "cos")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("acos")
    b.connect("clicked", self._operate, "acos")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("tan")
    b.connect("clicked", self._operate, "tan")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("atan")
    b.connect("clicked", self._operate, "atan")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("sinh")
    b.connect("clicked", self._operate, "sinh")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("asinh")
    b.connect("clicked", self._operate, "asinh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("cosh")
    b.connect("clicked", self._operate, "cosh")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("acosh")
    b.connect("clicked", self._operate, "acosh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("tanh")
    b.connect("clicked", self._operate, "tanh")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("atanh")
    b.connect("clicked", self._operate, "atanh")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    return t
  
  def make_exp(self):
    t = gtk.Table(2,5)
    x = 0
    
    b = gtk.Button("x^2")
    b.connect("clicked", self._operate, "square")
    t.attach(b, 0, 1, 0, 1)
    b = gtk.Button("sqrt")
    b.connect("clicked", self._operate, "sqrt")
    t.attach(b, 0, 1, 1, 2)
    x = x + 1
    
    b = gtk.Button("y^x")
    b.connect("clicked", self._operate, "pow")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("xroot")
    b.connect("clicked", self._operate, "xroot")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("log")
    b.connect("clicked", self._operate, "log")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("10^x")
    b.connect("clicked", self._operate, "10pow")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("ln")
    b.connect("clicked", self._operate, "ln")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("e^x")
    b.connect("clicked", self._operate, "epow")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("1/x")
    b.connect("clicked", self._operate, "inv")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    return t
  
  def make_val(self):
    t = gtk.Table(2,4)
    
    return t
  
  def make_stat(self):
    t = gtk.Table(2,6)
    x = 0
    
    b = gtk.Button("!")
    b.connect("clicked", self._operate, "!")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("mod")
    b.connect("clicked", self._operate, "mod")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("perm")
    b.connect("clicked", self._operate, "perm")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("comb")
    b.connect("clicked", self._operate, "comb")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("rand")
    b.connect("clicked", self._operate, "rand")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("frand")
    b.connect("clicked", self._operate, "frand")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    # divmod?
    #b = gtk.Button("mod")
    #b.connect("clicked", self._operate, "%")
    #t.attach(b, x, x+1, 0, 1)
    #b = gtk.Button("sqrt")
    #b.connect("clicked", self._operate, "sqrt")
    #t.attach(b, x, x+1, 1, 2)
    #x = x + 1
    
    b = gtk.Button("max")
    b.connect("clicked", self._operate, "max")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("min")
    b.connect("clicked", self._operate, "min")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("abs")
    b.connect("clicked", self._operate, "abs")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("sign")
    b.connect("clicked", self._operate, "sign")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("ceil")
    b.connect("clicked", self._operate, "ceil")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("floor")
    b.connect("clicked", self._operate, "floor")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    return t
  
  def make_int(self):
    t = gtk.Table(2,8)
    x = 0
    
    b = gtk.Button("->base")
    b.connect("clicked", self._operate, "->base")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("base")
    b.connect("clicked", self._operate, "base")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("->int")
    b.connect("clicked", self._operate, "->int")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("->float")
    b.connect("clicked", self._operate, "->float")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("shl")
    b.connect("clicked", self._operate, "shl")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("shr")
    b.connect("clicked", self._operate, "shr")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    #b = gtk.Button("(rol)")
    #b.connect("clicked", self._operate, "rol")
    #t.attach(b, x, x+1, 0, 1)
    #b = gtk.Button("(ror)")
    #b.connect("clicked", self._operate, "ror")
    #t.attach(b, x, x+1, 1, 2)
    #x = x + 1

    b = gtk.Button("&")
    b.connect("clicked", self._operate, "&")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("|")
    b.connect("clicked", self._operate, "|")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = gtk.Button("^")
    b.connect("clicked", self._operate, "^")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("~")
    b.connect("clicked", self._operate, "~")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    b = gtk.Button("and")
    b.connect("clicked", self._operate, "and")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("or")
    b.connect("clicked", self._operate, "or")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1
    
    b = gtk.Button("xor")
    b.connect("clicked", self._operate, "xor")
    t.attach(b, x, x+1, 0, 1)
    b = gtk.Button("not")
    b.connect("clicked", self._operate, "not")
    t.attach(b, x, x+1, 1, 2)
    x = x + 1

    #b = gtk.Button("aton")
    #b.connect("clicked", self._operate, "aton")
    #t.attach(b, x, x+1, 0, 1)
    #b = gtk.Button("ntoa")
    #b.connect("clicked", self._operate, "ntoa")
    #t.attach(b, x, x+1, 1, 2)
    #x = x + 1

    return t
  
  def make_modes(self):
    t = gtk.Table(2,4)
    x = 0

    b = gtk.Button("precision")
    b.connect("clicked", self._operate, "precision")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = gtk.Button()
    text = "deg"
    if self.cfg.angle_unit in ["rad", "radians"]:
      text = "rad"
    self._degrad_label = gtk.Label(text)
    b.add(self._degrad_label)
    b.connect("clicked", self._degrad)
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = gtk.Button("base")
    b.connect("clicked", self._operate, "base")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    b = gtk.Button("commas")
    b.connect("clicked", self._operate, "commas")
    t.attach(b, x, x+1, 0, 1)
    x = x + 1

    return t

  def make_bottomhalf(self):
    """TODO: make bottom half collapsable, so user has option
             of extra screen space and external input only
    """
    tabs = []
    
    numpad = self._make_numpad()
    tabs.append((numpad, "num"))

    if self.cfg.enable_alpha:
      alphapad = self._make_alphapad()
      tabs.append((alphapad, "alpha"))

	
    if len(tabs) > 1:
      nb = gtk.Notebook()
      self.vbox.pack_start(nb, 0, 1)
      for t in tabs:
        nb.append_page(t[0], gtk.Label(t[1]))
    else:
      self.vbox.pack_start(tabs[0][0], 0, 1)
	
    #grapher = gtk.Frame()
    #nb.append_page(grapher, gtk.Label("graph"))
    pass
    
  def _make_numpad(self):
    hb = gtk.HBox()
    hb.set_homogeneous(0)
    hb.set_spacing(10)

    t0 = gtk.Table(3, 2)
    
    b = gtk.Button("undo")
    b.connect("clicked", self._operate, "undo")
    t0.attach(b, 1, 2, 0, 1)

    b = gtk.Button("swap")
    b.connect("clicked", self._operate, "swap")
    t0.attach(b, 0, 1, 0, 1)

    b = gtk.Button("pick")
    b.connect("clicked", self._pick)
    t0.attach(b, 1, 2, 1, 2)

    b = gtk.Button("roll")
    b.connect("clicked", self._operate, "roll")
    t0.attach(b, 0, 1, 1, 2)

    #b = gtk.Button("rolln")
    #b.connect("clicked", self._operate, "roll")
    #t0.attach(b, 1, 2, 2, 3)

    b = gtk.Button("edit")
    b.connect("clicked", self._operate, "edit")
    t0.attach(b, 1, 2, 2, 3)

    b = gtk.Button("clear")
    b.connect("clicked", self._operate, "clear")
    t0.attach(b, 0, 1, 2, 3)

    hb.add(t0)
    
    
    t = gtk.Table(4,3)
    t.set_homogeneous(1)
    
    b0 = gtk.Button("0")
    bdot = gtk.Button(".")
    b1 = gtk.Button("1")
    b2 = gtk.Button("2")
    b3 = gtk.Button("3")
    b4 = gtk.Button("4")
    b5 = gtk.Button("5")
    b6 = gtk.Button("6")
    b7 = gtk.Button("7")
    b8 = gtk.Button("8")
    b9 = gtk.Button("9")
    
    t.attach(bdot, 1, 2, 3, 4)
    t.attach(b0, 0, 1, 3, 4)
    t.attach(b1, 0, 1, 2, 3)
    t.attach(b2, 1, 2, 2, 3)
    t.attach(b3, 2, 3, 2, 3)
    t.attach(b4, 0, 1, 1, 2)
    t.attach(b5, 1, 2, 1, 2)
    t.attach(b6, 2, 3, 1, 2)
    t.attach(b7, 0, 1, 0, 1)
    t.attach(b8, 1, 2, 0, 1)
    t.attach(b9, 2, 3, 0, 1)

    bdot.connect("clicked", self._type_into_entry, ".")
    b0.connect("clicked", self._type_into_entry, "0")
    b1.connect("clicked", self._type_into_entry, "1")
    b2.connect("clicked", self._type_into_entry, "2")
    b3.connect("clicked", self._type_into_entry, "3")
    b4.connect("clicked", self._type_into_entry, "4")
    b5.connect("clicked", self._type_into_entry, "5")
    b6.connect("clicked", self._type_into_entry, "6")
    b7.connect("clicked", self._type_into_entry, "7")
    b8.connect("clicked", self._type_into_entry, "8")
    b9.connect("clicked", self._type_into_entry, "9")

    bneg = gtk.Button("+/-")
    bneg.connect("clicked", self._operate, "neg")
    t.attach(bneg, 2, 3, 3, 4)

    hb.pack_start(t, 1, 1)

    t2 = gtk.Table(4,2)
    t2.set_homogeneous(1)

    #t2.attach(gtk.Button(), 0, 1, 0, 1)
    #t2.attach(gtk.Button(), 1, 2, 0, 1)
    bdrop = gtk.Button("<--")
    bdrop.connect("clicked", self._drop)
    t2.attach(bdrop, 0, 2, 0, 1)
    bplus = gtk.Button("+")
    bplus.connect("clicked", self._operate, "+")
    t2.attach(bplus, 0, 1, 1, 2)
    bminus = gtk.Button("-")
    bminus.connect("clicked", self._operate, "-")
    t2.attach(bminus, 0, 1, 2, 3)
    bmult = gtk.Button("*")
    bmult.connect("clicked", self._operate, "*")
    t2.attach(bmult, 1, 2, 1, 2)
    bdiv = gtk.Button("/")
    bdiv.connect("clicked", self._operate, "/")
    t2.attach(bdiv, 1, 2, 2, 3)
    bret = gtk.Button("Enter")
    bret.connect("clicked", self._run_entry)
    t2.attach(bret, 0, 2, 3, 4)


    hb.add(t2)

    return hb

  def _make_alphapad(self):
    self._shifted = 0

    rows = self.cfg.keymap
    wid = len(rows[0])
    hgt = len(rows)
    
    t = gtk.Table(hgt,wid)
    t.set_homogeneous(1)

    y = 0
    for row in rows:
      x = 0
      for l in row:
        if l != " ":
          b = gtk.Button(l)
          b.connect("clicked", self._type, l)
          t.attach(b, x, x+1, y, y+1)
        x = x + 1
      y = y + 1

    b = gtk.Button(" ")
    b.connect("clicked", self._type, " ")
    bpos = [b] ; bpos.extend(self.cfg.space_1_pos)
    apply(t.attach, bpos)
    b = gtk.Button(" ")
    b.connect("clicked", self._type, " ")
    bpos = [b] ; bpos.extend(self.cfg.space_2_pos)
    apply(t.attach, bpos)

    self.shift_label1 = gtk.Label("shf")
    self.shift_label2 = gtk.Label("shf")
    b = gtk.Button()
    b.add(self.shift_label1)
    b.connect("clicked", self._type, "shift")
    bpos = [b] ; bpos.extend(self.cfg.shift_1_pos)
    apply(t.attach, bpos)
    b = gtk.Button()
    b.add(self.shift_label2)
    b.connect("clicked", self._type, "shift")
    bpos = [b] ; bpos.extend(self.cfg.shift_2_pos)
    apply(t.attach, bpos)
    
    b = gtk.Button("<--")
    b.connect("clicked", self._drop)
    bpos = [b] ; bpos.extend(self.cfg.bkspc_pos)
    apply(t.attach, bpos)
    
    b = gtk.Button("TAB")
    b.connect("clicked", self._type, "\t")
    bpos = [b] ; bpos.extend(self.cfg.tab_pos)
    apply(t.attach, bpos)

    b = gtk.Button("RET")
    b.connect("clicked", self._run_entry)
    bpos = [b] ; bpos.extend(self.cfg.ret_pos)
    apply(t.attach, bpos)

    return t

  def _pick(self, b, *args):
    """Make the picked number match the visible stack...
       If user puts number on stack and picks, just run pick.
       If user enters number into entry, and presses pick,
       add one to it and then run pick.
    """
    t = self.entry.get_text()
    if t != "":
      self.entry.activate()
      self._operate(b, 1)
      self._operate(b, "+")

    self._operate(b, "pick")
  
  def _operate(self, b, func):
    """Automatically "press enter" if entry is not empty"""
    t = self.entry.get_text()
    if t != "":
      self.entry.activate()

    self.stack.operate(func)
    self.update_stack()

  def _type_into_entry(self, b, text):
    t = self.entry.get_text()
    p = self.entry.get_position()
    t = "%s%s%s" % (t[:p], text, t[p:])
    self.entry.set_text(t)
    self.entry.set_position(p+len(text))
    self.entry.grab_focus()

  def _type(self, b, text):
    if self._shifted:
      caps = self.cfg.caps
      if caps.has_key(text):
        text = caps[text]

    if len(text) > 1:
      if text == "shift":
        if self._shifted == 0:
          self._shift(1)
        elif self._shifted == 1:
          self._shift(2)
        else:
          self._shift(0)
    else:
      self._type_into_entry(b, text)
      if self._shifted == 1:
        self._shift(0)

  def _shift(self, val):
    self._shifted = val
    if val == 0:
      self.shift_label1.set_text("shf")
      self.shift_label2.set_text("shf")
    elif val == 1:
      self.shift_label1.set_text("Shf")
      self.shift_label2.set_text("Shf")
    elif val > 1:
      self.shift_label1.set_text("SHF")
      self.shift_label2.set_text("SHF")

  def _run_entry(self, *args):
    text = self.entry.get_text()
    if not text:  text = "dup"

    # Attempt to convert into a non-text datatype
    try:
      if "." in text:
        text = float(text)
      else:
        text = long(text, self.cfg.base)
    except:
      pass

    # Execute command
    self.stack.operate(text)

    self.entry.set_text("")
    self.update_stack()

  def _entry_backspace(self, *args):
    """TODO: make this behave better at beginning/end of line"""
    t = self.entry.get_text()
    p = self.entry.get_position()
    if t != ""  and  p > 0:
      t = t[:p-1] + t[p:]
      self.entry.set_text(t)
      self.entry.set_position(p-1)
      self.entry.grab_focus()

  def _degrad(self, b, *args):
    unit = "degrees"
    l = self._degrad_label
    l.set_text("deg")
    if self.cfg.angle_unit == "degrees":
      unit = "radians"
      l.set_text("rad")
    self._operate(None, unit)
    self._operate(None, "angle_unit")

  def _drop(self, *args):
    t = self.entry.get_text()
    if t != "":
      self._entry_backspace()
    else:
      self._operate(None, "drop")

  def about_screen(self, *args):
    text = \
"""%s V%s

Copyright (C) 2002 Scott Scriven <%s@xyzz.org>

This program is licened under the terms of the GNU General Public License (GPL), version 2.0 or later, as published by the Free Software Foundation.  See <http://www.gnu.org/copyleft/gpl.txt> for the terms of the latest version of the GNU General Public License.
""" % (program_name, program_version, program_name)

    ButtonChoiceWindow()("About", text, ["Ok"], long=1, transient=self.win)
    
  def functions_screen(self, *args):
    funcs = self.stack.functions.keys()
    funcs.sort()
    text = \
"""Function List
--------

You may execute commands either by clicking on buttons or by typing the name of the command into the stack.  Anything not recognized as a function will be pushed onto the stack.

The following %s functions are available:
""" % (len(funcs))
    for f in funcs:
      text = text + "\t%s\n" % (f)

    ButtonChoiceWindow()("Functions", text, ["Ok"], long=1, transient=self.win)
    
  def run(self):
    self.win.show_all()
    self.update_stack()
    gtk.mainloop()

  def destroy(self, *args):
    self.win.hide()
    self.win.unrealize()
    gtk.mainquit()

def run(args):
  global cfg
  #import config
  import pycfg
  cfg = pycfg.config(program_name)
  cfg.load()

  c = calculator(cfg)
  c.run()

if __name__ == "__main__":
  import sys
  run(sys.argv[1:])
